/*
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * File created by null on 7/5/09.
 */

#import "HCCController.h"
#import "HDLMaximumLengthTextFieldFormatter.h"
#import "NSAlertCheckbox.h"

// networking stuff
#import <unistd.h>
#import <errno.h>
#import <sys/time.h>
#import <sys/types.h>
#import <sys/socket.h>
#import <netinet/in.h>
#import <arpa/inet.h>
#import <netdb.h>
#import <fcntl.h>

#define MAX_BUFFER_SIZE				256
#define MAX_PASSWORD_LENGTH			10
#define MAX_HALO_SERVER_NAME_LENGTH	31
#define MAX_HALO_PASSWORD_LENGTH	8

#define APP_SUPPORT_PATH			[[NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES) objectAtIndex:0] stringByAppendingString:@"/Application Support/HaloClientChanger"]
#define SERVER_ADDRESSES_PATH		[APP_SUPPORT_PATH stringByAppendingString:@"/server_addresses.hcchistory"]
#define SERVER_PASSWORDS_PATH		[APP_SUPPORT_PATH stringByAppendingString:@"/server_passwords.hcchistory"]

NSString *HCCShouldShowRestartingHaloServerWarning = @"HCCShouldShowRestartingHaloServerWarning";
NSString *HCCShouldMakeHostSpectatorWarning = @"HCCShouldMakingHostSpectatorWarning";
NSString *HCCClientDisconnectNotification = @"HCCClientDisconnectNotification";

@implementation HCCController

#pragma mark Birth

+ (void)initialize
{
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES]
																						forKey:HCCShouldShowRestartingHaloServerWarning]];
	
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES]
																						forKey:HCCShouldMakeHostSpectatorWarning]];
}

- (void)awakeFromNib
{
	[mainWindow setExcludedFromWindowsMenu:YES];
	
	[serverAddressComboBox setDelegate:self];
	if ([[NSFileManager defaultManager] fileExistsAtPath:SERVER_ADDRESSES_PATH] && [[NSFileManager defaultManager] fileExistsAtPath:SERVER_PASSWORDS_PATH])
	{
		serverPasswordsHistoryArray = [[NSKeyedUnarchiver unarchiveObjectWithFile:SERVER_PASSWORDS_PATH] retain];
		serverAddressesHistoryArray = [[NSKeyedUnarchiver unarchiveObjectWithFile:SERVER_ADDRESSES_PATH] retain];
		[serverAddressComboBox addItemsWithObjectValues:serverAddressesHistoryArray];
		[serverAddressComboBox selectItemAtIndex:0];
	}
	
	[passwordTextField setFormatter:[HDLMaximumLengthTextFieldFormatter formatterWithMaximumLength:MAX_PASSWORD_LENGTH]];
	
	[haloServerNameTextField setFormatter:[HDLMaximumLengthTextFieldFormatter formatterWithMaximumLength:MAX_HALO_SERVER_NAME_LENGTH]];
	[haloPasswordTextField setFormatter:[HDLMaximumLengthTextFieldFormatter formatterWithMaximumLength:MAX_HALO_PASSWORD_LENGTH]];
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(haloMaxPlayersTextDidChange:)
												 name:NSTextDidChangeNotification
											   object:nil];
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(clientRequestedDisconnection:)
												 name:HCCClientDisconnectNotification
											   object:nil];
}

#pragma mark Connecting

- (IBAction)showConnectToServerWindow:(id)sender
{
	[connectToServerWindow makeKeyAndOrderFront:nil];
}

- (IBAction)showMainWindow:(id)sender
{
	[mainWindow makeKeyAndOrderFront:nil];
}

- (void)clientRequestedDisconnection:(NSNotification *)notification
{
	[self connectToServer:nil];
}

- (IBAction)connectToServer:(id)sender
{
	if ([[serverAddressComboBox stringValue] isEqualToString:@""])
	{
		NSBeep();
		NSRunAlertPanel(@"Invalid Server Address",
						@"You must enter in a server address to connect to.",
						@"OK", nil, nil);
	}
	else
	{
		if (!clientIsConnected)
		{
			[connectButton setEnabled:NO];
			[connectToServerProgressIndicator startAnimation:nil];
			[NSThread detachNewThreadSelector:@selector(connectToServerAndReceiveMessages:)
									 toTarget:self
								   withObject:nil];
		}
		else
		{
			clientIsConnected = NO;
		}
	}
}

- (void)comboBoxSelectionDidChange:(NSNotification *)notification
{
	if (serverAddressesHistoryArray && serverPasswordsHistoryArray && [serverAddressComboBox numberOfItems] > 0)
	{
		NSString *password = [serverPasswordsHistoryArray objectAtIndex:[serverAddressesHistoryArray indexOfObject:[serverAddressComboBox objectValueOfSelectedItem]]];
		if (password)
		{
			[passwordTextField setStringValue:password];
		}
	}
}

#pragma mark Receiving

- (void)connectToServerAndReceiveMessages:(id)object
{
	NSAutoreleasePool *connectionPool = [[NSAutoreleasePool alloc] init];
	
	socketDescriptor = socket(AF_INET, SOCK_STREAM, 0);
	
	if (socketDescriptor == -1)
	{
		perror("socket");
		
		NSBeep();
		NSRunAlertPanel(@"Connection Error",
						@"HClientChanger failed to initiate a socket connection.",
						@"OK", nil, nil);
		
		[connectButton setEnabled:YES];
		[connectToServerProgressIndicator stopAnimation:nil];
		[connectionPool release];
		[NSThread exit];
	}
	
	// find server address and port number
	NSArray *serverAddressComponents = [[serverAddressComboBox stringValue] componentsSeparatedByString:@":"];
	NSString *serverAddress = [serverAddressComponents objectAtIndex:0];
	int portNumber;
	if ([serverAddressComponents count] > 1)
	{
		portNumber = [[serverAddressComponents objectAtIndex:1] intValue];
	}
	else
	{
		const int DEFAULT_PORT = 2395;
		portNumber = DEFAULT_PORT;
	}
	
	struct hostent *hostEntry = gethostbyname([serverAddress UTF8String]);
	
	if (!hostEntry)
	{
		NSBeep();
		NSRunAlertPanel(@"Invalid Host Entry",
						@"HClientChanger failed to create a host entry for the server address you specified.",
						@"OK", nil, nil);
		
		close(socketDescriptor);
		[connectButton setEnabled:YES];
		[connectToServerProgressIndicator stopAnimation:nil];
		[connectionPool release];
		[NSThread exit];
	}
	
	struct sockaddr_in theirAddress;
	theirAddress.sin_family = AF_INET;
	theirAddress.sin_port = htons(portNumber);
	theirAddress.sin_addr.s_addr = inet_addr(inet_ntoa(*(((struct in_addr **)hostEntry->h_addr_list)[0])));
	memset(theirAddress.sin_zero, '\0', sizeof(theirAddress.sin_zero));
	
	// wait for five seconds to connect
	fcntl(socketDescriptor, F_SETFL, O_NONBLOCK);
	NSDate *startDate = [NSDate date];
	BOOL connectionDidSucceed = NO;
	
	while ([[NSDate date] timeIntervalSinceDate:startDate] < 5)
	{
		if (errno == EISCONN || connect(socketDescriptor, (struct sockaddr *)&theirAddress, sizeof(theirAddress)) == 0)
		{
			connectionDidSucceed = YES;
			break;
		}
	}
	
	[connectToServerProgressIndicator stopAnimation:nil];
	if (!connectionDidSucceed)
	{
		NSBeep();
		NSRunAlertPanel(@"No response from server",
						@"The server you are connecting to may be offline.",
						@"OK", nil, nil);
		
		close(socketDescriptor);
		[connectButton setEnabled:YES];
		[connectionPool release];
		[NSThread exit];
	}
	
	clientIsConnected = YES;
	
	NSString *passwordMessage = [@"pw" stringByAppendingFormat:@"%@\n", [passwordTextField stringValue]];
	send(socketDescriptor, [passwordMessage UTF8String], strlen([passwordMessage UTF8String]), 0);
	
	// read in connection message and wait for 5 seconds
	FD_ZERO(&masterSet);
	fd_set readSet = masterSet;
	FD_SET(socketDescriptor, &readSet);
	
	struct timeval timeValue;
	timeValue.tv_sec = 5;
	timeValue.tv_usec = 0;
	
	select(socketDescriptor + 1, &readSet, NULL, NULL, &timeValue);
	
	[connectButton setEnabled:YES];
	
	if (!FD_ISSET(socketDescriptor, &readSet))
	{
		NSLog(@"can't get password verification");
		NSBeep();
		NSRunAlertPanel(@"No response from server",
						@"The server you are connecting to may be offline.",
						@"OK", nil, nil);
		
		close(socketDescriptor);
		[connectionPool release];
		[NSThread exit];
	}
	
	[connectionPool release];

	int numberOfBytes = 0;
	BOOL serverIsValid = NO;
	
	do
	{
		NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
		
		if (!clientIsConnected)
		{
			break;
		}
		
		fd_set readSet = masterSet;
		FD_SET(socketDescriptor, &readSet);
		
		struct timeval timeValue;
		timeValue.tv_sec = 0;
		timeValue.tv_usec = 300000;
		
		select(socketDescriptor + 1, &readSet, NULL, NULL, &timeValue);
		
		if (!FD_ISSET(socketDescriptor, &readSet))
		{
			continue;
		}
		
		char buffer[MAX_BUFFER_SIZE];
		numberOfBytes = recv(socketDescriptor, buffer, MAX_BUFFER_SIZE, 0);
		
		if (numberOfBytes == -1)
		{
			[pool release];
			perror("recv");
			break;
		}
		else if (numberOfBytes > 0)
		{
			buffer[numberOfBytes] = '\0';
			
			NSArray *messageComponents = [[NSString stringWithUTF8String:buffer] componentsSeparatedByString:@"\n"];
			NSEnumerator *messageEnumerator = [messageComponents objectEnumerator];
			NSString *message;
			
			while (message = [messageEnumerator nextObject])
			{
				if (!message || [message length] < 2)
				{
					continue;
				}
				
				if ([message characterAtIndex:0] == 'g' && [message characterAtIndex:1] == 'd')
				{
					//NSLog(@"password accepted!");
					[connectToServerWindow orderOut:nil];
					[mainWindow makeKeyAndOrderFront:nil];
					
					serverIsValid = YES;
					
					if (!serverAddressesHistoryArray || !serverPasswordsHistoryArray)
					{
						serverAddressesHistoryArray = [[NSArray arrayWithObject:[serverAddressComboBox stringValue]] retain];
						serverPasswordsHistoryArray = [[NSArray arrayWithObject:[passwordTextField stringValue]] retain];
					}
					else
					{
						NSMutableArray *newServerAddressesArray = [NSMutableArray arrayWithArray:serverAddressesHistoryArray];
						NSMutableArray *newServerPasswordsArray = [NSMutableArray arrayWithArray:serverPasswordsHistoryArray];
						
						int indexToRemove = [newServerAddressesArray indexOfObject:[serverAddressComboBox stringValue]];
						if (indexToRemove != NSNotFound)
						{
							[newServerAddressesArray removeObjectAtIndex:indexToRemove];
							[newServerPasswordsArray removeObjectAtIndex:indexToRemove];
						}
						
						[newServerAddressesArray insertObject:[serverAddressComboBox stringValue]
													  atIndex:0];
						
						[newServerPasswordsArray insertObject:[passwordTextField stringValue]
													  atIndex:0];
						
						[serverAddressesHistoryArray release];
						[serverPasswordsHistoryArray release];
						
						serverAddressesHistoryArray = [newServerAddressesArray retain];
						serverPasswordsHistoryArray = [newServerPasswordsArray retain];
					}
					
					if (![[NSFileManager defaultManager] fileExistsAtPath:APP_SUPPORT_PATH])
					{
						[[NSFileManager defaultManager] createDirectoryAtPath:APP_SUPPORT_PATH
																   attributes:nil];
					}
					
					if (![NSKeyedArchiver archiveRootObject:serverAddressesHistoryArray
													 toFile:SERVER_ADDRESSES_PATH])
					{
						NSLog(@"Failure to save server_addresses.hcchistory");
					}
					
					if (![NSKeyedArchiver archiveRootObject:serverPasswordsHistoryArray
													 toFile:SERVER_PASSWORDS_PATH])
					{
						NSLog(@"Failure to save server_passwords.hcchistory");
					}
					
					[serverAddressComboBox removeAllItems];
					[serverAddressComboBox addItemsWithObjectValues:serverAddressesHistoryArray];
					[serverAddressComboBox selectItemAtIndex:0];
				}
				else if ([message characterAtIndex:0] == 'b' && [message characterAtIndex:1] == 'd')
				{
					NSRunAlertPanel(@"Bad password",
									@"The password you specified is incorrect.",
									@"OK", nil, nil);
					
					serverIsValid = YES;
					
					break;
				}
				else if ([message characterAtIndex:0] == 's' && [message characterAtIndex:1] == 'n')
				{
					// server name
					[haloServerNameTextField setStringValue:[message substringFromIndex:2]];
				}
				else if ([message characterAtIndex:0] == 'p' && [message characterAtIndex:1] == 'a')
				{
					// server password
					[haloPasswordTextField setStringValue:[message substringFromIndex:2]];
				}
				else if ([message characterAtIndex:0] == 'm' && [message characterAtIndex:1] == 'p')
				{
					// max players
					[haloMaxPlayersTextField setStringValue:[message substringFromIndex:2]];
					[haloMaxPlayersStepper setIntValue:[[message substringFromIndex:2] intValue]];
				}
				else if ([message characterAtIndex:0] == 's' && [message characterAtIndex:1] == 'u')
				{
					// halo server is up
					[applyButton setEnabled:YES];
					[restoreButton setEnabled:YES];
				}
				else if ([message characterAtIndex:0] == 's' && [message characterAtIndex:1] == 'd')
				{
					// halo server is down
					[applyButton setEnabled:NO];
					[restoreButton setEnabled:NO];
					[makeHostASpectatorMenuItem setState:NSOffState];
				}
				else if ([message characterAtIndex:0] == 'h' && [message characterAtIndex:1] == 'p')
				{
					// halo server restart progress
					[restartingHaloDemoProgressIndicator setDoubleValue:[[message substringFromIndex:2] intValue]];
					
					if (!restartingHaloDemoServerSession)
					{
						restartingHaloDemoServerSession = [NSApp beginModalSessionForWindow:restartingHaloDemoServerWindow];
					}
					
					[NSApp runModalSession:restartingHaloDemoServerSession];
				}
				else if ([message characterAtIndex:0] == 'h' && [message characterAtIndex:1] == 'e')
				{
					// halo server restart finished
					if (restartingHaloDemoServerSession)
					{
						[NSApp endModalSession:restartingHaloDemoServerSession];
						[restartingHaloDemoServerWindow orderOut:nil];
						restartingHaloDemoServerSession = NULL;
					}
				}
				else if ([message characterAtIndex:0] == 'h' && [message characterAtIndex:1] == 'c')
				{
					// halo server restart canceled
					[NSApp endModalSession:restartingHaloDemoServerSession];
					[restartingHaloDemoServerWindow orderOut:nil];
					restartingHaloDemoServerSession = NULL;
					
					NSBeep();
					NSRunAlertPanel(@"The host canceled your request",
									@"The Halo Demo server could not restart because the host canceled the request.",
									@"OK", nil, nil);
				}
				else if ([message characterAtIndex:0] == 'h' && [message characterAtIndex:1] == 'f')
				{
					// halo server failed to restart
					[NSApp endModalSession:restartingHaloDemoServerSession];
					[restartingHaloDemoServerWindow orderOut:nil];
					restartingHaloDemoServerSession = NULL;
					
					NSBeep();
					NSRunAlertPanel(@"Restart Failure",
									@"The Halo Demo server failed to restart.",
									@"OK", nil, nil);
				}
				else if ([message characterAtIndex:0] == 'h' && [message characterAtIndex:1] == 'a')
				{
					// halo server failed to restart with an error code
					[NSApp endModalSession:restartingHaloDemoServerSession];
					[restartingHaloDemoServerWindow orderOut:nil];
					restartingHaloDemoServerSession = NULL;
					
					NSBeep();
					NSRunAlertPanel(@"Restart Failure",
									@"Restarting Halo Demo has failed with an error code of %i.",
									@"OK", nil, nil, [[message substringFromIndex:2] intValue]);
				}
				else if ([message characterAtIndex:0] == 's' && [message characterAtIndex:1] == 's')
				{
					// spectating state
					[makeHostASpectatorMenuItem setState:[[message substringFromIndex:2] intValue]];
				}
			}
		}
		
		[pool release];
	}
	while (numberOfBytes != 0);
	
	close(socketDescriptor);
	
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	if (clientIsConnected)
	{
		if (!serverIsValid)
		{
			NSBeep();
			NSRunAlertPanel(@"Invalid Server",
							@"Either you are not connecting to a valid server or the server is not up.",
							@"OK", nil, nil);
		}
		else if (numberOfBytes == 0)
		{
			NSBeep();
			NSRunAlertPanel(@"Server Disconnection",
							@"Either you have disconnected from the server or the server has shut down.",
							@"OK", nil, nil);
		}
		
		clientIsConnected = NO;
	}
	else
	{
		[[NSNotificationCenter defaultCenter] postNotificationName:HCCClientDisconnectNotification
															object:self];
	}
	
	[applyButton setEnabled:NO];
	[restoreButton setEnabled:NO];
	
	[mainWindow orderOut:nil];
	[connectToServerWindow makeKeyAndOrderFront:nil];
	
	[pool release];
}

#pragma mark Sending

- (IBAction)restoreChanges:(id)sender
{
	send(socketDescriptor, "rs", 2, 0);
}

- (IBAction)applyChanges:(id)sender
{
	//[haloMaxPlayersTextField setIntValue:[haloMaxPlayersStepper intValue]];
	[haloMaxPlayersStepper setIntValue:[haloMaxPlayersTextField intValue]];
	
	NSString *message = [NSString stringWithFormat:@"sn%@\npa%@\nmp%@\n", [haloServerNameTextField stringValue], [haloPasswordTextField stringValue], [haloMaxPlayersTextField stringValue]];
	send(socketDescriptor, [message UTF8String], strlen([message UTF8String]), 0);
}

- (IBAction)restartHaloDemoServer:(id)sender
{
	BOOL userWantsToRestart = YES;
	if ([[NSUserDefaults standardUserDefaults] boolForKey:HCCShouldShowRestartingHaloServerWarning])
	{
		NSAlertCheckbox *alert = [NSAlertCheckbox alertWithMessageText:@"Do you want to continue?"
														 defaultButton:@"Yes"
													   alternateButton:@"No"
														   otherButton:nil
													   informativeText:@"This may only work with the standard gametypes Slayer and CTF. Other gametypes may not work unless their time limit is set at 15 minutes. This could cause harm on gametypes with a different time limit."];
		[alert setShowsCheckbox:YES];
		[alert setCheckboxText:@"Don't ask me again."];
		
		if ([alert runModal] == NSAlertAlternateReturn)
		{
			userWantsToRestart = NO;
		}
		else if ([alert checkboxState] == NSOnState)
		{
			[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithBool:NO]
													  forKey:HCCShouldShowRestartingHaloServerWarning];
		}
	}
	
	if (userWantsToRestart)
	{
		send(socketDescriptor, "hr\n", 3, 0);
	}
}

- (IBAction)cancelRestartingHaloDemoServer:(id)sender
{
	send(socketDescriptor, "hc\n", 3, 0);
	
	[NSApp endModalSession:restartingHaloDemoServerSession];
	[restartingHaloDemoServerWindow orderOut:nil];
	restartingHaloDemoServerSession = NULL;
}

- (IBAction)makeHostASpectator:(id)sender
{
	BOOL userWantsHostToSpectate = YES;
	
	if ([makeHostASpectatorMenuItem state] == NSOffState && [[NSUserDefaults standardUserDefaults] boolForKey:HCCShouldMakeHostSpectatorWarning])
	{
		NSAlertCheckbox *alert = [NSAlertCheckbox alertWithMessageText:@"Do you want to continue?"
														 defaultButton:@"Yes"
													   alternateButton:@"No"
														   otherButton:nil
													   informativeText:@"This will not work in Slayer and will only work in team based gametypes. Also, the host has to die once for every game to make this work."];
		[alert setShowsCheckbox:YES];
		[alert setCheckboxText:@"Don't ask me again."];
		
		if ([alert runModal] == NSAlertAlternateReturn)
		{
			userWantsHostToSpectate = NO;
		}
		else if ([alert checkboxState] == NSOnState)
		{
			[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithBool:NO]
													  forKey:HCCShouldMakeHostSpectatorWarning];
		}
	}
	
	if (userWantsHostToSpectate)
	{
		// tell server to toggle the spectator state
		send(socketDescriptor, "ss0\n", 3, 0);
	}
}

#pragma mark Other Stuff

- (IBAction)haloMaxPlayersStepper:(id)sender
{
	[haloMaxPlayersTextField setIntValue:[sender intValue]];
	[haloMaxPlayersTextField selectText:nil];
}

- (void)haloMaxPlayersTextDidChange:(NSNotification *)aNotification
{
	// we want the text field to be continuous
	
	if ([aNotification object] == [haloMaxPlayersTextField currentEditor])
	{
		if ([haloMaxPlayersTextField intValue] > 1 && [haloMaxPlayersTextField intValue] <= 16)
		{
			[haloMaxPlayersTextField setIntValue:[haloMaxPlayersTextField intValue]];
		}
	}
}

#pragma mark Menu Item Validation

- (BOOL)validateMenuItem:(NSMenuItem *)theMenuItem
{
	if ([theMenuItem action] == @selector(restartHaloDemoServer:))
	{
		return [applyButton isEnabled];
	}
	
	else if ([theMenuItem action] == @selector(makeHostASpectator:))
	{
		return [applyButton isEnabled];
	}
	
	else if ([theMenuItem action] == @selector(showMainWindow:))
	{
		[theMenuItem setState:[mainWindow isVisible]];
		return clientIsConnected;
	}
	
	return YES;
}

@end
