/*
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * File created by null on 7/5/09.
 */

#import "HSCPreferencesController.h"
#import "HSCController.h"
#import "HDLMaximumLengthTextFieldFormatter.h"

NSString *HSCListeningPort =		@"HSCListeningPort";
NSString *HSCServerPassword =		@"HSCServerPassword";
NSString *HSCEnableRemoteAccess =	@"HSCEnableRemoteAccess";

@implementation HSCPreferencesController

#pragma mark Birth

- (void)awakeFromNib
{	
	const int MAXIMUM_PASSWORD_LENGTH = 10;
	[serverPasswordTextField setFormatter:[HDLMaximumLengthTextFieldFormatter formatterWithMaximumLength:MAXIMUM_PASSWORD_LENGTH]];
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(serverShutOff:)
												 name:HSCServerShutOffNotification
											   object:nil];
	
}

- (void)showWindow
{
	[listeningPortTextField setIntValue:[[NSUserDefaults standardUserDefaults] integerForKey:HSCListeningPort]];
	[serverPasswordTextField setStringValue:[[NSUserDefaults standardUserDefaults] objectForKey:HSCServerPassword]];
	
	int enableRemoteAccessState = [[NSUserDefaults standardUserDefaults] integerForKey:HSCEnableRemoteAccess];
	[enableRemoteAccessButton setState:enableRemoteAccessState];
	[listeningPortTextField setEnabled:enableRemoteAccessState];
	[serverPasswordTextField setEnabled:enableRemoteAccessState];
	
	if (enableRemoteAccessState == NSOffState)
	{
		[serverPasswordTextFieldLabel setTextColor:[NSColor disabledControlTextColor]];
		[listeningPortTextFieldLabel setTextColor:[NSColor disabledControlTextColor]];
	}
	else
	{
		[serverPasswordTextFieldLabel setTextColor:[NSColor controlTextColor]];
		[listeningPortTextFieldLabel setTextColor:[NSColor controlTextColor]];
	}
	
	[window makeKeyAndOrderFront:nil];
}

#pragma mark Actions

- (IBAction)enableRemoteAccess:(id)sender
{
	[serverPasswordTextField setEnabled:[sender state]];
	[listeningPortTextField setEnabled:[sender state]];
	
	if ([sender state] == NSOnState)
	{
		[window makeFirstResponder:serverPasswordTextField];
		[serverPasswordTextFieldLabel setTextColor:[NSColor controlTextColor]];
		[listeningPortTextFieldLabel setTextColor:[NSColor controlTextColor]];
	}
	else
	{
		[serverPasswordTextFieldLabel setTextColor:[NSColor disabledControlTextColor]];
		[listeningPortTextFieldLabel setTextColor:[NSColor disabledControlTextColor]];
	}
}

- (IBAction)applyChanges:(id)sender
{
	[[HSCController sharedController] enableRemoteAccess:[enableRemoteAccessButton state]];
	[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:[enableRemoteAccessButton state]]
											  forKey:HSCEnableRemoteAccess];
	
	if ([enableRemoteAccessButton state] == NSOnState && [[NSUserDefaults standardUserDefaults] integerForKey:HSCListeningPort] != [listeningPortTextField intValue] || ![[[NSUserDefaults standardUserDefaults] objectForKey:HSCServerPassword] isEqualToString:[serverPasswordTextField stringValue]])
	{
		serverIsRestarting = YES;
	}
	
	[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:[listeningPortTextField intValue]]
											  forKey:HSCListeningPort];
	
	[[NSUserDefaults standardUserDefaults] setObject:[serverPasswordTextField stringValue]
											  forKey:HSCServerPassword];
	
	if (serverIsRestarting)
	{
		[[HSCController sharedController] enableRemoteAccess:NO];
		[applyingChangesProgressIndicator startAnimation:nil];
		[applyChangesButton setEnabled:NO];
	}
}

- (void)serverShutOff:(NSNotification *)notification
{
	if (serverIsRestarting)
	{
		[[HSCController sharedController] enableRemoteAccess:YES];
		[applyingChangesProgressIndicator stopAnimation:nil];
		[applyChangesButton setEnabled:YES];
		
		serverIsRestarting = NO;
	}
}

- (void)turnOffServer
{
	[enableRemoteAccessButton setState:NSOffState];
	[serverPasswordTextField setEnabled:NO];
	[listeningPortTextField setEnabled:NO];
	
	[serverPasswordTextFieldLabel setTextColor:[NSColor disabledControlTextColor]];
	[listeningPortTextFieldLabel setTextColor:[NSColor disabledControlTextColor]];
	
	[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:NO]
											  forKey:HSCEnableRemoteAccess];
}

@end
